{ =============================================================
(c) 2000 Andrei Kashin (Andrei.Kashin@pobox.spbu.ru)

Version 1.00

================================================================}

unit URLDownload;

interface

uses Windows, URLMon, ActiveX, Classes;

type

TDownloadProgress = record
 Progress: cardinal;
 ProgressMax: cardinal;
 StatusCode: cardinal;
 StatusText: string;
end;
PDownloadProgress = ^TDownloadProgress;

THTTPRequestData = record
 Method: string;
 ExtraHeaders: string;
 RequestBody: string;
end;
PHTTPRequestData = ^THTTPRequestData;

TURLDownloader = class;

TDownloadStartNotify =
 procedure (Sender: TURLDownloader) of object;
TDownloadProgressNotify =
 procedure (Sender: TURLDownloader; const Progress:TDownloadProgress) of object;
TBytesAvailableNotify =
 procedure (Sender: TURLDownloader; BytesAvailable: integer;
            var BytesToRead: integer; var ReadBuffer: pointer) of object;
TDataReadNotify =
 procedure (Sender: TURLDownloader; Data:pointer; DataSize: integer) of object;
TDownloadStopNotify =
 procedure (Sender: TURLDownloader; ErrorCode: cardinal) of object;
THTTPResponseNotify =
 procedure (Sender: TURLDownloader; ResponseCode: DWord;
            const ResponseHeaders: string) of object;


TURLDownloader = class (TInterfacedObject, IBindStatusCallback,
                        IHttpNegotiate, IServiceProvider)
 private
  FURL: string;
  FBinding: IBinding;
  FTotalRead: integer;
  FDownloadInProgress: boolean;
  FAutoDestroy: boolean;
  FDownloadAborted: boolean;
  FBindFlags: DWord;

  FOnDownloadStart: TDownloadStartNotify;
  FOnDownloadProgress: TDownloadProgressNotify;
  FOnBytesAvailable: TBytesAvailableNotify;
  FOnDataRead: TDataReadNotify;
  FOnDownloadStop: TDownloadStopNotify;
  FOnHTTPResponse: THTTPResponseNotify;

  FHTTPRequestData: THTTPRequestData;

  procedure SetBindFlags(Value: DWord);

  {IBindStatusCallback}
  function OnStartBinding(dwReserved: DWORD; pib: IBinding): HResult; stdcall;
  function GetPriority(out nPriority): HResult; stdcall;
  function OnLowResource(reserved: DWORD): HResult; stdcall;
  function OnProgress(ulProgress, ulProgressMax, ulStatusCode: ULONG;
                      szStatusText: LPCWSTR): HResult; stdcall;
  function OnStopBinding(hresult: HResult; szError: LPCWSTR): HResult; stdcall;
  function GetBindInfo(out grfBINDF: DWORD; var bindinfo: TBindInfo): HResult; stdcall;
  function OnDataAvailable(grfBSCF: DWORD; dwSize: DWORD; formatetc: PFormatEtc;
                           stgmed: PStgMedium): HResult; stdcall;
  function OnObjectAvailable(const iid: TGUID; punk: IUnknown): HResult; stdcall;
  {IHTTPNegotiate}
  function BeginningTransaction(szURL, szHeaders: LPCWSTR; dwReserved: DWORD;
      out szAdditionalHeaders: LPWSTR): HResult; stdcall;
  function OnResponse(dwResponseCode: DWORD; szResponseHeaders, szRequestHeaders: LPCWSTR;
      out szAdditionalRequestHeaders: LPWSTR): HResult; stdcall;
  {IServiceProvider}
  function QueryService(const rsid, iid: TGuid; out Obj): HResult; stdcall;
 protected
  procedure DoOnDataRead(Data:pointer; DataSize: integer); virtual;
  procedure DoOnBytesAvailable(BytesAvailable:integer; var BytesToRead: integer;
                               var ReadBuffer: pointer); virtual;
  procedure DoOnDownloadStop(ErrorCode: cardinal); virtual;
 public
  constructor Create;
  destructor Destroy; override;
  procedure BeforeDestruction; override;

  procedure StartDownload(Caller:IUnknown; const DownloadURL:string;
                          HTTPRequestData:PHTTPRequestData = nil);
  procedure AbortDownload;

  property URL:string read FURL;
  property DownloadInProgress: boolean read FDownloadInProgress;
  property AutoDestroy:boolean read FAutoDestroy write FAutoDestroy;
  property BindFlags: DWord read FBindFlags write SetBindFlags;
  property TotalRead: integer read FTotalRead;

  property OnDownloadStart: TDownloadStartNotify
            read FOnDownloadStart write FOnDownloadStart;
  property OnDownloadProgress: TDownloadProgressNotify
            read FOnDownloadProgress write FOnDownloadProgress;
  property OnBytesAvailable: TBytesAvailableNotify
            read FOnBytesAvailable write FOnBytesAvailable;
  property OnDataRead: TDataReadNotify
            read FOnDataRead write FOnDataRead;
  property OnDownloadStop: TDownloadStopNotify
            read FOnDownloadStop write FOnDownloadStop;
  property OnHTTPResponse: THTTPResponseNotify
            read FOnHTTPResponse write FOnHTTPResponse;
end;

TDownloaderToStream = class;

TDownloadToStreamNotify =
 procedure (Sender: TDownloaderToStream; ErrorCode: cardinal;
            Stream: TStream; StartPosition: integer) of object;

TDownloaderToStream = class(TURLDownloader)
 private
  FStream: TStream;
  FStartPos: integer;
  FMemoryStreamCreated: boolean;
  FOnDownloadToStream: TDownloadToStreamNotify;
 protected
  procedure DoOnDataRead(Data:pointer; DataSize: integer); override;
  procedure DoOnBytesAvailable(BytesAvailable:integer; var BytesToRead: integer;
                               var ReadBuffer: pointer); override;
  procedure DoOnDownloadStop(ErrorCode: cardinal); override;

  property OnBytesAvailable;
  property OnDataRead;
  property OnDownloadStop;
 public
  constructor Create(Stream:TStream);
  destructor Destroy; override;

  property OnDownloadToStream: TDownloadToStreamNotify
             read FOnDownloadToStream write FOnDownloadToStream;
end;

implementation

uses SysUtils;

const MaxReadBufferSize = 1024 * 64; // 64 kb

function StringToWideChar(const s:string):LPWSTR;
var ResultLen: integer;
begin
 ResultLen := MultiByteToWideChar(0, 0, Pointer(s), -1, nil, 0);
 Result := CoTaskMemAlloc(ResultLen shl 1);
 MultiByteToWideChar(0, 0, Pointer(s), -1, Result, ResultLen);
end;

procedure ClearWideChar(WideChar:PWideChar);
begin
 CoTaskMemFree(WideChar);
end;


{ TURLDownloader }

procedure TURLDownloader.AbortDownload;
begin
 if FBinding <> nil then begin
  FDownloadAborted := true;
  FBinding.Abort;
 end;
end;

procedure TURLDownloader.BeforeDestruction;
begin
 dec(FRefCount); // compensate _AddRef in constructor
 try
  inherited BeforeDestruction;
 except
 end; 
end;

constructor TURLDownloader.Create;
begin
 inherited Create;
 _AddRef; // to prevent premature Destroy
 FBindFlags := BINDF_ASYNCHRONOUS;
end;

destructor TURLDownloader.Destroy;
begin
 FOnDownloadStop := nil;
 FAutoDestroy := false;
 AbortDownload;
 inherited Destroy;
end;

procedure TURLDownloader.DoOnBytesAvailable(BytesAvailable: integer;
  var BytesToRead: integer; var ReadBuffer: pointer);
begin
 if Assigned(FOnBytesAvailable) then
  FOnBytesAvailable(Self, BytesAvailable, BytesToRead, ReadBuffer);
end;

procedure TURLDownloader.DoOnDataRead(Data: pointer; DataSize: integer);
begin
 if Assigned(FOnDataRead) then
  FOnDataRead(Self, Data, DataSize);
end;

procedure TURLDownloader.DoOnDownloadStop(ErrorCode: cardinal);
begin
 if Assigned(FOnDownloadStop) then
  FOnDownloadStop(Self, ErrorCode);
end;

function TURLDownloader.GetBindInfo(out grfBINDF: DWORD;
                                    var bindinfo: TBindInfo): HResult;
//
var P: pointer;
begin
 grfBINDF := FBindFlags;

 if FHTTPRequestData.Method <> '' then begin
  if CompareText(FHTTPRequestData.Method, 'GET') = 0 then
   bindinfo.dwBindVerb := BINDVERB_GET
  else if CompareText(FHTTPRequestData.Method, 'POST') = 0 then
   bindinfo.dwBindVerb := BINDVERB_POST
  else if CompareText(FHTTPRequestData.Method, 'PUT') = 0 then
   bindinfo.dwBindVerb := BINDVERB_PUT
  else begin
   bindinfo.dwBindVerb := BINDVERB_CUSTOM;
   bindinfo.szCustomVerb := StringToWideChar(FHTTPRequestData.Method);
  end;
 end
 else
  bindinfo.dwBindVerb := BINDVERB_GET;

 if (FHTTPRequestData.RequestBody <> '') and
    (bindinfo.dwBindVerb <> BINDVERB_GET)
 then begin
  bindinfo.stgmedData.hGlobal :=
   GlobalAlloc(GMEM_SHARE, Length(FHTTPRequestData.RequestBody));
  P := GlobalLock(bindinfo.stgmedData.hGlobal);
  if P <> nil then
   Move(PChar(FHTTPRequestData.RequestBody)^, P^, Length(FHTTPRequestData.RequestBody));
  GlobalUnlock(bindinfo.stgmedData.hGlobal);
  bindinfo.stgmedData.tymed := TYMED_HGLOBAL;
  bindinfo.cbstgmedData := Length(FHTTPRequestData.RequestBody);
 end;

 Result := S_OK;
end;

function TURLDownloader.GetPriority(out nPriority): HResult;
begin
 Result := E_NOTIMPL;
end;

function TURLDownloader.OnDataAvailable(grfBSCF, dwSize: DWORD;
  formatetc: PFormatEtc; stgmed: PStgMedium): HResult;
//
var BytesRead: integer;
    ReadBuffer: pointer;
    BufferSize: integer;
    BufferAllocated: boolean;
    BytesAvailable: integer;
    tmpRes: HResult;
begin
 BytesAvailable := integer(dwSize) - FTotalRead;
 BufferAllocated := false;

 try //except
  try //finally
   repeat
    ReadBuffer := nil;
    BufferSize := BytesAvailable;
    if BytesAvailable > MaxReadBufferSize then
     BufferSize := MaxReadBufferSize;
    DoOnBytesAvailable(BytesAvailable, BufferSize, ReadBuffer);
    if FDownloadAborted or (BufferSize < 0) then begin
     Result := E_Abort;
     exit;
    end;
    if (BufferSize = 0) or (BufferSize > BytesAvailable) then begin
     if grfBSCF and BSCF_LASTDATANOTIFICATION <> 0 then
      Result := E_FAIL
     else
      Result := S_OK;
     exit;
    end;
    if ReadBuffer = nil then begin
     GetMem(ReadBuffer, BufferSize);
     BufferAllocated := true;
    end;

    tmpRes := IStream(stgmed^.stm).Read(ReadBuffer, BufferSize, @BytesRead);
    if ((tmpRes = S_OK) or (DWord(tmpRes) = E_PENDING)) and (BytesRead = BufferSize)
    then begin
     DoOnDataRead(ReadBuffer, BufferSize);
     dec(BytesAvailable, BufferSize);
     inc(FTotalRead, BufferSize);
     if FDownloadAborted then begin
      Result := E_Abort;
      exit;
     end;
    end
    else begin
     Result := E_FAIL;
     exit;
    end;

    if BufferAllocated then begin
     FreeMem(ReadBuffer);
     BufferAllocated := false;
    end;
   until BytesAvailable = 0;

  finally
   if BufferAllocated then FreeMem(ReadBuffer);
  end;
  Result := S_OK;
 except
  Result := E_FAIL;
 end;
end;

function TURLDownloader.OnLowResource(reserved: DWORD): HResult;
begin
 Result := E_NOTIMPL;
end;

function TURLDownloader.OnObjectAvailable(const iid: TGUID; punk: IUnknown): HResult;
begin
 Result := E_NOTIMPL;
end;

function TURLDownloader.OnProgress(ulProgress, ulProgressMax,
  ulStatusCode: ULONG; szStatusText: LPCWSTR): HResult;
//
var Progress: TDownloadProgress;
begin
 if Assigned(FOnDownloadProgress) then begin
  Progress.Progress := ulProgress;
  Progress.ProgressMax := ulProgressMax;
  Progress.StatusCode := ulStatusCode;
  Progress.StatusText := WideCharToString(szStatusText);
  try
   FOnDownloadProgress(Self, Progress);
  except
   // don't leak out exceptions
  end;
 end;
 Result := S_OK;
end;

function TURLDownloader.OnStartBinding(dwReserved: DWORD; pib: IBinding): HResult;
begin
 FBinding := pib;
 Result := S_OK;
end;

function TURLDownloader.OnStopBinding(hresult: HResult; szError: LPCWSTR): HResult;
var ErrorCode: DWord;
    szResult: PWideChar;
    clsidProtocol: TGUID;
begin
 if hResult <> S_OK then begin
  szResult := nil;
  FBinding.GetBindResult(clsidProtocol, ErrorCode, szResult, 0);
  if szResult <> nil then CoTaskMemFree(szResult);
 end
 else
  ErrorCode := S_OK;
 FBinding := nil;
 Finalize(FHTTPRequestData);
 try
  DoOnDownloadStop(ErrorCode);
 except
  // don't leak out exceptions
 end;
 FDownloadInProgress := false;
 FDownloadAborted := false;
 if FAutoDestroy then Free;
 Result := S_OK;
end;

{IHTTPNegotiate}
function TURLDownloader.BeginningTransaction(szURL, szHeaders: LPCWSTR;
  dwReserved: DWORD; out szAdditionalHeaders: LPWSTR): HResult;
begin
 if FHTTPRequestData.ExtraHeaders <> '' then
  szAdditionalHeaders := StringToWideChar(FHTTPRequestData.ExtraHeaders);
 Result := S_OK;
end;

function TURLDownloader.OnResponse(dwResponseCode: DWORD;
  szResponseHeaders, szRequestHeaders: LPCWSTR;
  out szAdditionalRequestHeaders: LPWSTR): HResult;
begin
 if Assigned(FOnHTTPResponse) then
  try
   FOnHTTPResponse(Self, dwResponseCode, WideCharToString(szResponseHeaders));
  except
   // don't leak out exceptions
  end;
 Result := S_OK;
end;

const
 IID_IStream: TGuid = '{0000000C-0000-0000-C000-000000000046}';

procedure TURLDownloader.StartDownload(Caller:IUnknown; const DownloadURL:string;
                                       HTTPRequestData:PHTTPRequestData = nil);
var BindingHost: UrlMon.IBindHost;
    ClientSite: IOleClientSite;
    BindContext: IBindCtx;
    Moniker: IMoniker;
    Stream: pointer;
    PrevBindStatusCallBack: IBindStatusCallBack;
    URLW: PWideChar;
    FullURLW: PWideChar;
begin
 if FDownloadInProgress then exit; //? or raise an exception
 FDownloadInProgress := true;
 FURL := DownloadURL;
 FTotalRead := 0;
 if Assigned(FOnDownloadStart) then FOnDownloadStart(Self);
 if HTTPRequestData <> nil then
  FHTTPRequestData := HTTPRequestData^;
 URLW := StringToWideChar(FURL);

 try
  if Caller <> nil then begin
   if (Caller as IOleObject).GetClientSite(ClientSite) <> S_OK then Abort;
   BindingHost := ClientSite as UrlMon.IBindHost;
   if (CreateAsyncBindCtx(0, Self as IBindStatusCallBack, nil, BindContext) <> S_OK) or
      (BindingHost.CreateMoniker(URLW, BindContext, Moniker, 0) <> S_OK) or
      Failed(BindingHost.MonikerBindToStorage(Moniker, BindContext, Self,
                                              IID_IStream, Stream))
   then Abort;
   if Stream <> nil then IStream(Stream)._Release;
   FullURLW := nil;
   if Moniker.GetDisplayName(BindContext, nil, FullURLW) = S_OK then begin
    FURL := WideCharToString(FullURLW);
    ClearWideChar(FullURLW);
   end;
  end
  else begin
   if (CreateURLMoniker(nil, URLW, Moniker) <> S_OK) or
      (CreateAsyncBindCtx(0, Self as IBindStatusCallback, nil, BindContext) <> S_OK) or
      (RegisterBindStatusCallback(BindContext, Self, PrevBindStatusCallBack, 0) <> S_OK) or
      Failed(Moniker.BindToStorage(BindContext, nil, IID_IStream, Stream))
   then Abort;
   if Stream <> nil then IStream(Stream)._Release;
  end;
 except
  if FDownloadInProgress then begin
   Finalize(FHTTPRequestData);
   try
    DoOnDownloadStop(DWord(E_FAIL));
   except
    // don't leak out exceptions
   end;
   FDownloadInProgress := false;
  end;
 end;
 ClearWideChar(URLW);
end;

function TURLDownloader.QueryService(const rsid, iid: TGuid; out Obj): HResult;
begin
 Result := QueryInterface(rsid, Obj);
end;

procedure TURLDownloader.SetBindFlags(Value: DWord);
begin
 FBindFlags := (Value or BINDF_ASYNCHRONOUS) and not BINDF_ASYNCSTORAGE;
end;

{ TDownloaderToStream }

constructor TDownloaderToStream.Create(Stream: TStream);
begin
 inherited Create;
 if Stream = nil then begin
  FStream := TMemoryStream.Create;
  FMemoryStreamCreated := true;
 end
 else FStream := Stream;
 FStartPos := FStream.Position;
end;

destructor TDownloaderToStream.Destroy;
begin
 inherited Destroy;
 if FMemoryStreamCreated then FStream.Free;
end;

procedure TDownloaderToStream.DoOnBytesAvailable(BytesAvailable: integer;
  var BytesToRead: integer; var ReadBuffer: pointer);
begin
 // do nothing
end;

procedure TDownloaderToStream.DoOnDataRead(Data: pointer; DataSize: integer);
begin
 FStream.Write(Data^, DataSize);
end;

procedure TDownloaderToStream.DoOnDownloadStop(ErrorCode: cardinal);
begin
 if Assigned(FOnDownloadToStream) then
  FOnDownloadToStream(Self, ErrorCode, FStream, FStartPos);
 if FMemoryStreamCreated then TMemoryStream(FStream).Clear;
end;

end.
